# Copyright (c) 2014 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from wsgiref import simple_server


def application(_, start_response):
    response = "crtauth client version too old. Update your client\n"
    status = '403 Forbidden'
    response_headers = [('Content-Type', 'text/plain'),
                        ('Content-Length', str(len(response)))]
    start_response(status, response_headers)
    return response,


if __name__ == '__main__':
    simple_server.make_server('', 8080, application).serve_forever()



# Copyright (c) 2015 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


def constant_time_compare(x, y):
    """
    Compares two byte strings in a way such that execution time is constant
    regardless of how much alike the input values are, provided that they
    are of the same length.

    Comparisons between user input and secret data such as calculated
    HMAC values needs to be executed in constant time to avoid leaking
    information to the caller via the timing side channel.

    Params:
        x: the first byte string to compare
        y: the second byte string to compare
    Return: True if x and y are equal, else False
    """
    if len(x) != len(y):
        return False
    n = reduce(lambda a, b: a | b, (ord(x) ^ ord(y) for x, y in zip(x, y)))
    return n == 0




# Copyright (c) 2011-2015 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import with_statement

import hashlib
import hmac
import logging
import re
import time

from crtauth import ssh
from crtauth import exceptions
from crtauth import protocol
from crtauth import msgpack_protocol

# Previously, create_response lived in this module. Importing to preserve
# backward compatibility for now. New users should use the version in
# the client module.
# noinspection PyUnresolvedReferences
from crtauth.client import create_response

# The maximum number of seconds from a challenge is created to the time when
# the generated response is processed. This needs to compensate for clock
# skew between servers in a cluster.
RESP_TIMEOUT = 20

# The number of seconds a clock can be off before we start getting too
# old / too new messages
CLOCK_FUDGE = 2

# A token has validity encoded in it. Tokens with far too long validity
# indicates that something is fishy, and we should reject them. This
# number is the maximum allowed token validity, in seconds.
MAX_TOKEN_LIFETIME = 600

log = logging.getLogger("crtauth.server")


class AuthServer(object):
    def __init__(self, secret, key_provider, server_name, token_lifetime=60,
                 now_func=time.time, lowest_supported_version=0):
        """
        Constructs a new AuthServer object, with the given parameters.

        @param secret a string containing the secret data used to authenticate
        requests

        @param key_provider a subclass of KeyProvider

        @param server_name a text string included in the challenge that needs
        to match the host that the client connects to

        @param token_lifetime the number of seconds that a token issued by
        this instance should be valid for

        @param now_func a function that returns current time in unix seconds.
        Used for testing

        @lowest_supported_version an integer that indicates the lowest protocol
        version that is supported, 0 represents the unversioned first protocol
        """

        self.token_lifetime = token_lifetime
        self.secret = secret
        self.key_provider = key_provider
        self.urandom = open("/dev/urandom", "r")
        if not re.match("^[a-zA-Z0-9.-]+$", server_name):
            raise ValueError(
                "Invalid server name, can only contain letters, numbers, "
                "dot (.) and hyphen (-): " + server_name)
        if len(server_name) > 255:
            raise ValueError("Too long length of server_name: " +
                             len(server_name))
        self.server_name = server_name
        self.now_func = now_func
        self.lowest_supported_version = lowest_supported_version

    def create_challenge(self, username, version=0):
        """This method returns a challenge suitable for ssh-agent signing.

        @param username the username of the user requesting a challenge
        @param version the highest protocol version the clients supports
        @exception ProtocolVersionError if the client supports
        """
        if len(username) > 64:
            raise ValueError("Username is too long: " + username)

        try:
            key = self.key_provider.get_key(username)
            fingerprint = key.fingerprint()
        except exceptions.NoSuchUserException:
            log.info("No public key found for '%s', faking it." % username)
            fingerprint = self._hmac(username)[:6]

        if version < 1:
            if self.lowest_supported_version > version:
                raise exceptions.ProtocolVersionError(
                    "Client needs to support at least version %d"
                    % self.lowest_supported_version
                )

            c = protocol.Challenge(fingerprint=fingerprint,
                                   server_name=self.server_name,
                                   unique_data=self.urandom.read(20),
                                   valid_from=int(self.now_func() - CLOCK_FUDGE),
                                   valid_to=int(self.now_func() + RESP_TIMEOUT),
                                   username=username)
            b = c.serialize()

            payload = protocol.VerifiablePayload(digest=self._hmac(b), payload=b)
            return ssh.base64url_encode(payload.serialize())
        else:
            c = msgpack_protocol.Challenge(
                fingerprint=fingerprint,
                server_name=self.server_name,
                unique_data=self.urandom.read(20),
                valid_from=int(self.now_func() - CLOCK_FUDGE),
                valid_to=int(self.now_func() + RESP_TIMEOUT),
                username=username)
            return ssh.base64url_encode(c.serialize(self.secret))

    def create_token(self, response):
        """
        This method verifies that the response given from the client
        is valid and if so returns a token used for authentication.
        """
        s = ssh.base64url_decode(response)

        if s[0] == 'r':
            # this is a version 0 response
            version_1 = False
            if self.lowest_supported_version > 0:
                raise exceptions.ProtocolVersionError(
                    "Client needs to support at least version %d"
                    % self.lowest_supported_version
                )
            r = protocol.Response.deserialize(s)
            if not r.hmac_challenge.verify(self._hmac):
                raise exceptions.InvalidInputException(
                    "Challenge hmac verification failed, not matching  secret"
                )
            challenge = protocol.Challenge.deserialize(r.hmac_challenge.payload)
        elif s[0] == '\x01':
            # this is a version 1 response
            version_1 = True
            r = msgpack_protocol.Response.deserialize(s)
            challenge = msgpack_protocol.Challenge.deserialize_authenticated(
                r.challenge, self.secret)
        else:
            raise exceptions.ProtocolError("invalid first byte of response")

        # verify the integrity of the challenge in the response
        if self.server_name != challenge.server_name:
            s = "Got challenge with the wrong server_name encoded"
            raise exceptions.InvalidInputException(s)

        key = self.key_provider.get_key(challenge.username)

        if challenge.valid_from > self.now_func():
            s = time.strftime("%Y-%m-%d %H:%M:%S UTC",
                              time.gmtime(challenge.valid_from))
            raise exceptions.InvalidInputException("Response with challenge "
                                                   "created as %s too new "
                                                   % s)

        if challenge.valid_to < self.now_func():
            s = time.strftime("%Y-%m-%d %H:%M:%S UTC",
                              time.gmtime(challenge.valid_from))
            raise exceptions.InvalidInputException("Response with challenge "
                                                   "created as %s too old "
                                                   % s)

        if version_1:
            if not key.verify_signature(r.signature, r.challenge):
                raise exceptions.InvalidInputException(
                    "Client did not provide proof that it controls "
                    "the secret key")
        else:
            if not key.verify_signature(r.signature, r.hmac_challenge.payload):
                raise exceptions.InvalidInputException(
                    "Client did not provide proof that it controls "
                    "the secret key")

        expire_time = int(self.now_func()) + self.token_lifetime

        return self._make_token(challenge.username, expire_time)

    def validate_token(self, token):
        buf = ssh.base64url_decode(token)
        hmac_token = protocol.VerifiablePayload.deserialize(buf)

        t = protocol.Token.deserialize(hmac_token.payload)

        if t.valid_to < self.now_func():
            s = "Token expired at " + time.strftime("%Y-%m-%d %H:%M:%S UTC",
                                                    time.gmtime(t.valid_to))
            raise exceptions.TokenExpiredException(s)

        if t.valid_from > self.now_func():
            s = time.strftime("%Y-%m-%d %H:%M:%S UTC",
                              time.gmtime(t.valid_from))
            raise exceptions.TokenExpiredException("Token created at %s" % s)

        if (t.valid_to - t.valid_from) > MAX_TOKEN_LIFETIME:
            raise exceptions.InvalidInputException("Token lifetime too long")

        if not hmac_token.verify(self._hmac):
            raise exceptions.InvalidInputException("Token hmac verification "
                                                   "failed, not matching our "
                                                   "secret")

        return t.username

    def _hmac(self, data):
        return hmac.new(self.secret, data, hashlib.sha1).digest()

    def _make_token(self, username, expire_time):
        t = protocol.Token(username=username,
                           valid_from=int(self.now_func() - CLOCK_FUDGE),
                           valid_to=expire_time)

        b = t.serialize()

        payload = protocol.VerifiablePayload(digest=self._hmac(b), payload=b)

        return ssh.base64url_encode(payload.serialize())



# Copyright (c) 2011-2013 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
from crtauth import exceptions
from crtauth import rsa


class KeyProvider(object):
    """Provides PubKey instances based on username. Please note that the
    username provided is not sanity checked and may be provided by an
    untrusted user. """

    def get_key(self, username):
        """Get a PubKey instance from some underlying system"""
        raise NotImplementedError


class FileKeyProvider(KeyProvider):
    def __init__(self, dir):
        self.dir = dir

    def get_key(self, username):
        if "/" in username:
            raise exceptions.CrtAuthError("Don't trick me into opening files "
                                          "by having slash in username!")
        fn = "%s/%s_id_rsa.pub" % (self.dir, username)
        if not os.path.exists(fn):
            raise exceptions.NoSuchUserException()
        with open(fn, "r") as f:
            return rsa.RSAPublicKey(f.read())



# Copyright (c) 2011-2013 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import base64
import sys
import os
import socket
import struct

from crtauth import exceptions
from crtauth import rsa

# ssh-agent communication protocol constants
SSH2_AGENTC_REQUEST_IDENTITIES = 11
SSH2_AGENT_IDENTITIES_ANSWER = 12
SSH2_AGENTC_SIGN_REQUEST = 13
SSH2_AGENT_SIGN_RESPONSE = 14


def base64url_decode(s):
    """
    Decodes a url-safe base64 encoded string.
    """

    if type(s) is unicode:
        s = s.encode("US-ASCII")

    if len(s) % 4 == 3:
        s += "="
    elif len(s) % 4 == 2:
        s += "=="

    # b64decode is real crap when checking for the validity of the input.
    try:
        return base64.b64decode(s, "-_")
    except:
        _, e, tb = sys.exc_info()
        raise exceptions.InvalidInputException(
            "Invalid base64 sequence: %s" % e), None, tb


def base64url_encode(data):
    """
    encodes a url-safe base64 encoded string.
    """

    s = base64.b64encode(data, "-_")
    return s.rstrip("=")


def i2s(i):
    return struct.pack("!I", i)


def write_field(socket, data):
    if type(data) is int:
        data = chr(data)
    elif type(data) is not str:
        data = repr(data)
    socket.send(i2s(len(data)) + data)


class SigningPlug(object):
    """
    Base-class for signing a challenge with an RSA-key. This class
    should never be used directly. The actual signing capabilities are
    implemented in it's subclasses.

    Signing classes can also be used as context managers.
    """
    def sign(self, data, fingerprint):
        raise NotImplementedError("Don't use the SigningPlug directly. This "
                                  "is an abstract base class")

    def __enter__(self):
        pass

    def __exit__(self, exc_class, exc_value, bt):
        self.close()

    def close(self):
        pass


class SingleKeySigner(SigningPlug):
    """Simple implementation of SigningPlug, which holds a single private
    key for signing.

    SingleKeySigner is intended for services, cron-jobs and similar that does
    not have an ssh-agent readily available.
    """
    def __init__(self, priv_key):
        self.key = rsa.RSAPrivateKey(priv_key)

    def sign(self, data, fingerprint):
        return self.key.sign(data)


class AgentSigner(SigningPlug):
    """
    An implementation of SigningPlug, which gets it's private key from an
    ssh-agent.

    AgentSigner is intended for command line tools where their invoker
    also controls an ssh-agent process that can be contacted via a UNIX
    referenced by the SSH_AUTH_SOCK environment variable.
    """
    def __init__(self):
        self.sock = socket.socket(socket.AF_UNIX)
        sock_path = os.getenv("SSH_AUTH_SOCK")
        if not sock_path:
            raise exceptions.SshAgentError(
                "The environment variable SSH_AUTH_SOCK is not set. Please "
                "configure your ssh-agent.")
        self.sock.settimeout(5.0)
        self.sock.connect(sock_path)

    def __find_key(self, key_fingerprint):
        write_field(self.sock, SSH2_AGENTC_REQUEST_IDENTITIES)
        length, response_code, count = struct.unpack("!IBI",
                                                     self.sock.recv(9))
        assert response_code == SSH2_AGENT_IDENTITIES_ANSWER
        resp = self.sock.recv(length - 5)
        fields = rsa.read_fields(resp)
        for i in xrange(count):
            try:
                key = rsa.RSAPublicKey(fields.next())
            except exceptions.KeyError:
                fields.next()
                continue
            fields.next()  # ignore filename for key
            if key.fingerprint() == key_fingerprint:
                return key

    def sign(self, data, fingerprint):
        try:
            pub_key = self.__find_key(fingerprint)

            if not pub_key:
                raise exceptions.SshAgentError(
                    "Your ssh-agent does not have the required key added. This "
                    "usually indicates that ssh-add has not been run.")

            self.sock.send(i2s(len(pub_key) + len(data) + 13) +
                           chr(SSH2_AGENTC_SIGN_REQUEST))
            write_field(self.sock, pub_key)
            write_field(self.sock, data)
            self.sock.send("\0\0\0\0")
            length, response_code, resp_len = struct.unpack("!IBI",
                                                            self.sock.recv(9))
            assert response_code == SSH2_AGENT_SIGN_RESPONSE
            buf = self.sock.recv(length - 5)
            fields = rsa.read_fields(buf)
            response_type = fields.next()
            assert response_type == "ssh-rsa"
            return fields.next()
        except socket.timeout as why:
            raise exceptions.SshAgentError(why)

    def close(self):
        self.sock.close()



# Copyright (c) 2011-2013 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import xdrlib

from crtauth import exceptions
from crtauth.constant_time_compare import constant_time_compare

# to support another serialization mechanism, replace these with references
# to other classes that implements the needed methods for all handled data
# types
packer_class = xdrlib.Packer
unpacker_class = xdrlib.Unpacker


class Field(object):
    """
    Instances of the subclasses of Field knows how to pack and unpack
    themselves into and out of a [un]packer instance that provide the methods
    of xdrlib.Packer() and xdrlib.Unpacker()
    """

    def pack(self, packer, value):
        """
        Pack a value using the specified packer.

        :param packer: Active packer to append value to.
        :param value: Value to pack.
        """
        raise NotImplementedError("pack")

    def unpack(self, unpacker):
        """
        Unpack a value using the specified unpacker.

        :param unpacker: Active unpacker to extract value from.
        """
        raise NotImplementedError("unpack")


class FString(Field):
    """
    Fixed length string type.
    """
    def __init__(self, size):
        self.size = size

    def pack(self, packer, value):
        return packer.pack_fstring(self.size, value)

    def unpack(self, unpacker):
        return unpacker.unpack_fstring(self.size)


class String(Field):
    """
    Variable length string type.
    """
    def pack(self, packer, value):
        return packer.pack_string(value)

    def unpack(self, unpacker):
        return unpacker.unpack_string()


class UInt(Field):
    """
    Unsigned integer.
    """
    def pack(self, packer, value):
        return packer.pack_uint(value)

    def unpack(self, unpacker):
        return unpacker.unpack_uint()


class Type(Field):
    """
    Encapsulate other type.
    """
    def __init__(self, cls):
        self.cls = cls

    def pack(self, packer, value):
        return packer.pack_string(value.serialize())

    def unpack(self, unpacker):
        return self.cls.deserialize(unpacker.unpack_string())


class SerializablePacket(object):
    __magic__ = None
    __fields__ = None

    def __init__(self, **kw):
        if len(kw) != len(self.__fields__):
            raise exceptions.ProtocolError("Field length mismatch")

        for key, _ in self.__fields__:
            val = kw.get(key, None)
            if val is None:
                raise exceptions.ProtocolError("Missing required argument "
                                               "'%s'" % key)
            setattr(self, key, val)

    def serialize(self):
        if self.__magic__ is None or self.__fields__ is None:
            raise exceptions.ProtocolError(
                "Serialization can only be performed on classes implementing "
                "__fields__ and __magic__")

        p = packer_class()

        p.pack_fstring(1, self.__magic__)

        for name, field in self.__fields__:
            value = getattr(self, name)
            field.pack(p, value)

        return p.get_buffer()

    @classmethod
    def deserialize(cls, buf):
        if cls.__magic__ is None or cls.__fields__ is None:
            raise exceptions.ProtocolError(
                "Deserialization can only be performed on classes "
                "implementing __fields__ and __magic__")

        u = unpacker_class(buf)

        if u.unpack_fstring(1) != cls.__magic__:
            raise exceptions.ProtocolError(
                "Wrong magic byte for " + cls.__name__ +
                " (should be '" + hex(ord(cls.__magic__)) + "')")

        kw = dict()

        for name, field in cls.__fields__:
            kw[name] = field.unpack(u)

        return cls(**kw)


class VerifiablePayload(SerializablePacket):
    """
    A digest and payload combination.

    digest - The payload digest.
    payload - The payload.
    """

    __magic__ = 'v'

    __fields__ = [
        ("digest", FString(20)),
        ("payload", String()),
    ]

    def verify(self, digest_f):
        return constant_time_compare(self.digest, digest_f(self.payload))


class Challenge(SerializablePacket):
    """
    Represents a challenge, with binary serialization support
    """

    __magic__ = 'c'

    __fields__ = [
        ("unique_data", FString(20)),
        ("valid_from", UInt()),
        ("valid_to", UInt()),
        ("fingerprint", String()),
        ("server_name", String()),
        ("username", String()),
    ]


class Response(SerializablePacket):
    """
    Represents a response to a challenge.

    signature - A signature generated by the client.
    hmac_challenge - A verifiable packed challenge (see VerifiablePayload).
    """
    __magic__ = 'r'

    __fields__ = [
        ("signature", String()),
        ("hmac_challenge", Type(VerifiablePayload))
    ]


class Token(SerializablePacket):
    """
    A token which verifies that can be used to authorize a specific user.
    """
    __magic__ = 't'

    __fields__ = [
        ("valid_from", UInt()),
        ("valid_to", UInt()),
        ("username", String()),
    ]



# Copyright (c) 2013-2014 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import sys
import logging
from crtauth import ssh

log = logging.getLogger("crtauth.wsgi")


class CrtauthMiddleware(object):
    """
    An instance of this class acts as middleware that will handle crtauth
    HTTP authentication for connecting clients. Once a user is authenticated,
    the WSGI environ object will have a value for the key 'crtauth.username'
    that corresponds to the username of the authenticated user.

    By default, only properly authenticated requests will reach the wrapped
    WSGI application. If you want more fine grained control over which
    resources should be protected, use the manual_authorization constructor
    parameter which will hand over responsibility for returning the
    401 Unauthorized HTTP status to the wrapped application.

    Any exception thrown by the auth server will be logged and result in a
    "403 Forbidden" response.
    """
    CHAP_HEADER = "X-CHAP"
    CHAP_WSGI_HEADER = "HTTP_X_CHAP"
    AUTHORIZATION_HEADER = "HTTP_AUTHORIZATION"
    AUTHORIZATION_TYPE = "chap"

    STATUS_UNAUTHORIZED = '401 Unauthorized'
    STATUS_FORBIDDEN = '403 Forbidden'
    STATUS_INTERNAL_SERVER_ERROR = '500 Internal Server Error'
    STATUS_OK = '200 OK'
    
    CHAP_PATH = "/_auth"

    AUTH_ENVIRON = "crtauth.username"

    def __init__(self, app, auth_server, disabled=False,
                 manual_authorization=False):
        """
        Construct a WSGI middleware that authenticates it's users before
        passing on requests to the enclosed app.

        :param app: a WSGI application to wrap.
        :param auth_server: A configured crtauth.server.AuthServer instance.
        :param disabled: set to true if authentication is disabled.
        :param manual_authorization: set to true if this middleware should
         not require the user to be authenticated to pass along the request.
         In this case the app needs to take care to return status 401 when
         an unauthorized user attempts to access a protected resource unless
         the WSGI environ key 'crtauth.username' maps to an acceptable
         username.
        """

        self.app = app
        self.auth_server = auth_server
        self.disabled = disabled
        self.manual_authorization = manual_authorization

    def read_authorization(self, environ):
        return environ.get(self.AUTHORIZATION_HEADER, None)

    def __call__(self, environ, start_response):
        """
        Handle the request and make sure that CHAP authentication stages for the
        special path /_auth

        Also check for any existing outstanding authentication by reading the
        Authorization header for a token matching "chap:<token>".
        This token is validated using the auth_server associated with this
        middleware.

        environ - Request environment.
        start_response - WSGI invokable start_response.
        """
        # if disabled, skip authorization part.
        if self.disabled:
            return self.app(environ, start_response)

        return self.handle_handshake(environ, start_response)

    def handle_handshake(self, environ, start_response):
        if environ.get("PATH_INFO", "/") == self.CHAP_PATH:
            try:
                return self.handshake_path(environ, start_response)
            except:
                log.error("Error when handling request", exc_info=sys.exc_info())
                start_response(self.STATUS_INTERNAL_SERVER_ERROR, [])
                return []

        return self.handle_authorization(environ, start_response)

    def handle_authorization(self, environ, start_response):
        username = None
        authorization = self.read_authorization(environ)

        if authorization and authorization.startswith("chap:"):
            _, token = authorization.split(":", 1)

            try:
                username = self.auth_server.validate_token(token)
                environ.update({self.AUTH_ENVIRON: username})
            except:
                log.warning("Failed to validate token", exc_info=sys.exc_info())
                return self.handle_unauthorized(environ, start_response)

        if not username and not self.manual_authorization:
            return self.handle_unauthorized(environ, start_response)

        return self.app(environ, start_response)

    def handshake_path(self, environ, start_response):
        """
        Client is in the handshake phase.

        Expect that the X-CHAP header is set, otherwise raise 403 Forbidden.

        Perform the required step as specified for the X-CHAP header, or fail with
        handle_auth_server_exception.
        """
        chap_header = environ.get(self.CHAP_WSGI_HEADER, None)

        if chap_header is None:
            start_response(self.STATUS_FORBIDDEN, [])
            return []

        method, value = chap_header.split(":", 1)

        if method == "request":
            return self.handle_request(environ, start_response, value)

        if method == "response":
            return self.handle_response(environ, start_response, value)

        log.warning("Unknown chap method: " + method)
        return self.handle_auth_server_exception(environ, start_response)

    def handle_request(self, environ, start_response, request):
        username, version = self.parse_request(request)
        try:
            challenge = self.auth_server.create_challenge(username, version)
        except:
            log.warning("Failed to create challenge", exc_info=sys.exc_info())
            return self.handle_auth_server_exception(environ, start_response)

        start_response(self.STATUS_OK, [(self.CHAP_HEADER, "challenge:" + challenge)])
        return []

    def handle_response(self, environ, start_response, response):
        try:
            token = self.auth_server.create_token(response)
        except:
            log.warning("Failed to create token", exc_info=sys.exc_info())
            return self.handle_auth_server_exception(environ, start_response)

        start_response(self.STATUS_OK, [(self.CHAP_HEADER, "token:" + token)])
        return []

    def handle_auth_server_exception(self, environ, start_response):
        """
        Raise a "403 Forbidden" error to the client.
        """
        start_response(self.STATUS_FORBIDDEN, [])
        return ["Failed to authenticate user"]

    def handle_unauthorized(self, environ, start_response):
        start_response(self.STATUS_UNAUTHORIZED,
                       [("content-type", "text/plain")])
        return ["Unauthorized"]

    @staticmethod
    def parse_request(request):
        """
        This method contains logic to detect a v1 and beyond request and
        differentiate it from a version 0 request, which is just an ascii
        username. While all v1 requests are also valid usernames (the curse
        and blessing of base64 encoding) it is pretty unlikely that a username
        happens to also decode to a valid msgpack message with the correct
        magic values.

        @return a tuple containing username then version
        """
        binary = ssh.base64url_decode(request)
        if len(binary) < 4:
            return request, 0
        if ord(binary[0]) > 4 or binary[1] != 'q':
            # This code handles version values up to 4. Should give plenty
            # of time to forget all about the unversioned version 0
            return request, 0
        b = ord(binary[2])
        if (b < 0xa1 or b > 0xbf) and b != 0xd9:
            # third byte does not indicate a string longer than 0 and shorter
            # than 256 octets long (According to UTF_8 rfc3629, a unicode
            # char can encode to at most 4 UTF-8 bytes, and username values
            # in crtauth is limited to 64 characters, thus the max number of
            # bytes a username can be is 64 * 4 == 256
            return request, 0
        if b == 0xd9:
            username_start = 4
            username_len = ord(binary[3])
        else:
            username_start = 3
            username_len = ord(binary[2]) & 0x1f

        if len(binary) - username_start < username_len:
            # decoded string is too short
            return request, 0

        return binary[username_start:username_start + username_len], 1



# Copyright (c) 2011-2015 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# A standalone implementation of RSA signature generation and verification
# as used in SSH with SHA-1 as hash algorithm, specified in RFC3447 section
# 8.2.

import base64
import hashlib
import binascii

from crtauth import exceptions
from crtauth.constant_time_compare import constant_time_compare


class RSAPrivateKey(object):
    def __init__(self, private_key):
        """private_key is expected to be in PEM format as described in
        RFC3447 A.1.2. This is what ssh-keygen outputs in the id_rsa output
        file"""
        private_key = private_key.strip()
        decoded = base64.b64decode("".join(private_key.split("\n")[1:-1]))

        items = _read_items(decoded)

        self.private_exp = items[3]
        self.mod = items[1]
        # it turns out that ssh writes leading zeroes, which we get rid of
        # by doing a round trip to a (very long) int.
        self.mod_size = len(_int_to_str(self.mod))
        self.padding = _make_padding(self.mod_size)

    def encrypt(self, data):
        if len(data) > self.mod_size:
            raise exceptions.KeyError("Key size too small, more than %d bytes "
                                      "of data can not be encrypted" %
                                      self.mod_size)
        return _int_to_str(pow(_str_to_int(data), self.private_exp, self.mod))

    def sign(self, data):
        digest = hashlib.sha1(data).digest()
        return self.encrypt(self.padding + digest)


class RSAPublicKey(object):
    """Instances of this class represents a public RSA key"""

    def __init__(self, key):
        """@param key the ASCII string from id_rsa.pub from ssh-keygen"""
        if key.startswith("ssh-rsa"):
            self.decoded = key.split(" ")[1]
            self.encoded = base64.b64decode(self.decoded)
        else:
            self.encoded = key
            self.decoded = base64.b64encode(self.encoded)

        self.fp = hashlib.sha1(self.encoded).digest()[:6]

        fields = read_fields(self.encoded)
        sigtype = fields.next()
        if sigtype != "ssh-rsa":
            raise exceptions.KeyError("Unknown key type %s. This code "
                                      "currently only supports ssh-rsa" %
                                      sigtype)
        self.exp = _str_to_int(fields.next())
        self.mod = _str_to_int(fields.next())
        # it turns out that ssh writes leading zeroes, which we get rid of
        # by roundtripping to bignum.
        self.mod_size = len(_int_to_str(self.mod))

    def __repr__(self):
        return self.encoded

    def __len__(self):
        return len(self.encoded)

    def fingerprint(self):
        return self.fp

    def decrypt(self, cleartext):
        n = _str_to_int(cleartext)
        return _int_to_str(pow(n, self.exp, self.mod))

    def verify_signature(self, signature, data):
        # for some reason, highest byte in padding is 0 that disappears in
        # the encryption roundtrip, so we need to add zeroes to be able to
        # compare.
        decrypted = self.decrypt(signature)
        if len(decrypted) < self.mod_size:
            decrypted = ("\x00" *
                         (self.mod_size - len(decrypted))) + decrypted
        padded_digest = (_make_padding(self.mod_size) +
                         hashlib.sha1(data).digest())
        return constant_time_compare(padded_digest, decrypted)


def read_fields(bytes):
    off = 0
    while off < len(bytes):
        l = s2i(bytes[off:])
        off += 4
        yield bytes[off:off + l]
        off += l


def s2i(data):
    """Read four bytes off the provided byte string and return the value as
    a big endian 32 bit unsigned integer"""
    num = 0
    for i, val in enumerate(data[:4]):
        num += ord(val) << ((3 - i) * 8)
    return num


def _make_padding(mod_length):
    """
    Creates a padding string that when concatenated with the SHA-1 checksum
    results in a EMSA-PKCS1-v1_5 encoded string as specified in RFC3447
    section 9.2
    """

    # Constant for SHA-1 taken from RFC3447 page 42 note 1
    PS = "\x30\x21\x30\x09\x06\x05\x2b\x0e\x03\x02\x1a\x05\x00\x04\x14"
    all_ff = "\xff" * (mod_length - len(PS) - 23)
    return "\x00\x01" + all_ff + "\x00" + PS


def _read_items(data):
    """
    Very simplistic ASN.1 parser, assuming the top level is a COLLECTION and
    the contents of that some primitive types.
    """
    offset = 0

    items = []
    while len(data) > offset:
        type = ord(data[offset])
        offset += 1
        length = 0
        if ord(data[offset]) & 0x80:
            len_octets = ord(data[offset]) ^ 0x80
            offset += 1
            for i in range(len_octets):
                length += ord(data[offset + i]) << (len_octets - i - 1) * 8
            offset += len_octets
        else:
            length = ord(data[offset])
            offset += 1
        if type == 0x30:
            return _read_items(data[offset:offset + length])
        elif type == 0x02:
            items.append(_str_to_int(data[offset:offset + length]))

        offset += length

    return items


def _str_to_int(data):
    """Treats the provided string as a sequence of octets and interprets
    the octets as an arbitrarily sized unsigned integer in network byte
    order
    """
    return int(data.encode('hex'), 16)


def _int_to_str(num):
    """Packs the num (which must be convertible to a long) into a
       byte string. The resulting byte string is the big-endian two's
       complement representation of the passed in long."""
    s = hex(num)[2:]
    s = s.rstrip('L')
    if len(s) & 1:
        s = '0' + s
    s = binascii.unhexlify(s)
    return s



# Copyright (c) 2011-2014 Spotify AB
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from crtauth import exceptions, rsa, key_provider
import ldap
from ldap import filter


class LDAPKeyProvider(key_provider.KeyProvider):
    """
    Provides a PubKey instance based on a lookup in an LDAP directory.

    User entries are expected to be of class posixAccount living directly under
    base_dn in the directory information tree, have an uid matching the
    username parameter and one sshPublicKey string representation
    of the ssh public key of the user.

    Group entries are expected to be of class posixGroup and be located under
    cn=groups under the base_dn in the directory information tree. The group
    string parameter corresponds to the cn attribute of the posixGroup entry
    """
    def __init__(self, uri, auth_user, auth_password, base_dn, group=None):
        """
        Constructs and binds an LDAPKeyProvider instance to the server
        identified by the uri using auth_user and auth_password for
        authentication.

        When users are looked up, it is verified that they belong to the
        provided group.
        """
        self.group = None
        if group:
            self.group = filter.escape_filter_chars(group)
        self.base_dn = base_dn

        # I know, this is not functionality the ldap module straightforwardly
        # exposes, but it seems to work.
        self.conn = ldap.ldapobject.ReconnectLDAPObject(uri)
        self.conn.simple_bind(auth_user, auth_password)

    def get_key(self, username):
        """
        Returns a PubKey instance based on LDAP lookup. If group is specified
        in the constructor, the user needs to be a member for the lookup to
        succeed.

        Throws NoSuchUserException, InsufficientPrivilegesException,
        MissingKeyException when appropriate.
        """

        user = filter.escape_filter_chars(username)
        f = ("(|(&(uid=%s)(objectClass=posixAccount))"
             "(&(memberUid=%s)(objectClass=posixGroup)))" % (user, user))

        # We don't care about looking for a group if self.group is not set
        group_dn = None
        if self.group:
            group_dn = "cn=%s,cn=groups,%s" % (self.group, self.base_dn)

        result = dict(self.conn.search_s(self.base_dn, ldap.SCOPE_SUBTREE, f,
                                         ['sshPublicKey']))

        attributes = result.get("uid=%s,cn=users,%s" % (user, self.base_dn))
        if attributes is None:
            raise exceptions.NoSuchUserException("User '%s' not found" % user)

        key_list = attributes.get("sshPublicKey")
        if key_list is None:
            raise exceptions.MissingKeyException("User '%s' does not have "
                                                 "her key in LDAP" % user)
        if len(key_list) > 1:
            raise RuntimeError("Can't handle multiple sshPublicKey values "
                               "for an LDAP user")

        if group_dn and group_dn not in result:
            s = ("User '%s' not member of required group '%s'" %
                 (user, self.group))
            raise exceptions.InsufficientPrivilegesException(s)

        return rsa.RSAPublicKey(key_list[0])


class HybridKeyProvider(key_provider.KeyProvider):
    """
    A KeyProvider that behaves as an LDAP KeyProvider if there is no ldap data
    it falls back to a FileKeyProvider.
    Useful for non mixing real ldap users with service-specific non-human
    users.
    """

    def __init__(self, dir, uri, auth_user, auth_password, base_dn, group=None):
        """
        Constructs a FileKeyProvider based on the directory dir, and a
        LDAPKeyProvider based on the remaining arguments.
        """
        self.file_key_provider = key_provider.FileKeyProvider(dir)
        self.ldap_key_provider = LDAPKeyProvider(uri, auth_user, auth_password,
                                                 base_dn, group)

    def get_key(self, username):
        """
        Returns the user's public key if it can be found in LDAP, otherwise
        tries to find it in the key directory, or fails.
        """
        try:
            return self.ldap_key_provider.get_key(username)
        except exceptions.NoSuchUserException:
            try:
                return self.file_key_provider.get_key(username)
            except Exception, e:
                raise exceptions.NoSuchUserException(
                    "User %s not in ldap, defaulted to pubkey dir and got "
                    "exception %s" % (username, e))
